/*
 * Written by Hanson Char and released to the public domain,
 * as explained at http://creativecommons.org/licenses/publicdomain
 */
package net.sf.beanlib.util.concurrent;

import java.util.AbstractQueue;
import java.util.Collection;
import java.util.Iterator;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.LockSupport;

/**
 * An unbounded concurrent blocking queue implemented upon 
 * {@link java.util.concurrent.ConcurrentLinkedQueue ConcurrentLinkedQueue}
 * and intrinsic locking for queueing consumers.
 * <p>
 * Note there is currently no such class in Java 6.
 * <p>
 * In contrast to {@link java.util.concurrent.LinkedBlockingQueue LinkedBlockingQueue}
 * which is always bounded, a ConcurrentLinkedBlockingQueue is unbounded.
 * 
 * @author Hanson Char
 * @param <E> the type of elements held in this collection
 */
public class SynchronousConcurrentLinkedBlockingQueue<E> extends AbstractQueue<E>
        implements java.io.Serializable, BlockingQueue<E>
{
    private static final long serialVersionUID = 8153969967701180306L;

    private final AtomicReference<Thread> parked = new AtomicReference<Thread>();
    
    private final ConcurrentLinkedQueue<E> q;

    public SynchronousConcurrentLinkedBlockingQueue() {
        q = new ConcurrentLinkedQueue<E>();
    }

    public SynchronousConcurrentLinkedBlockingQueue(Collection<? extends E> c) {
        q = new ConcurrentLinkedQueue<E>(c);
    }

    @Override
    public Iterator<E> iterator() {
        return q.iterator();
    }

    @Override
    public int size() {
        return q.size();
    }

    public boolean offer(E e) {
        q.offer(e);
        Thread t = parked.get();
        
        if (t != null) {
            if (parked.compareAndSet(t, null))
                LockSupport.unpark(t);
        }
        return true;
    }

    public E peek() {
        return q.peek();
    }

    public E poll() {
        return q.poll();
    }

    /**
     * Retrieves and removes the head of this queue, waiting if necessary until
     * an element becomes available.
     * 
     * @return the head of this queue
     * @throws InterruptedException if interrupted while waiting
     */
    public E take() throws InterruptedException 
    {
        for (;;) {
            E e = q.poll();

            if (e != null)
                return e;
            if (Thread.interrupted())
                throw new InterruptedException();
            synchronized(this) {
                parked.set(Thread.currentThread());
                e = q.poll();
            
                if (e != null) {
                    parked.compareAndSet(Thread.currentThread(), null);
                    return e;
                }
                LockSupport.park();
                parked.compareAndSet(Thread.currentThread(), null);
            }
            if (Thread.interrupted())
                throw new InterruptedException();
        }
    }
    
    /**
     * Retrieves and removes the head of this queue, waiting up to the specified
     * wait time if necessary for an element to become available.
     * 
     * @param timeout
     *            how long to wait before giving up, in units of <tt>unit</tt>.
     *            A negative timeout is treated the same as to wait forever.
     * @param unit
     *            a <tt>TimeUnit</tt> determining how to interpret the
     *            <tt>timeout</tt> parameter
     * @return the head of this queue, or <tt>null</tt> if the specified
     *         waiting time elapses before an element is available
     * @throws InterruptedException if interrupted while waiting
     */
    public E poll(final long timeout, final TimeUnit unit) throws InterruptedException 
    {
        if (timeout < 0)
            return take();  // treat -ve timeout same as to wait forever
        final long t1 = System.nanoTime() + unit.toNanos(timeout);
        
        for (;;) {
            E e = q.poll();

            if (e != null)
                return e;
            final long duration = t1 - System.nanoTime();
            
            if (duration <= 0)
                return null;    // time out
            if (Thread.interrupted())   
                throw new InterruptedException();
            synchronized(this) {
                parked.set(Thread.currentThread());
                e = q.poll();
            
                if (e != null) {
                    parked.compareAndSet(Thread.currentThread(), null);
                    return e;
                }
                LockSupport.parkNanos(duration);
                parked.compareAndSet(Thread.currentThread(), null);
            }
            if (Thread.interrupted()) 
                throw new InterruptedException();
        }
    }
    
    public void put(E e) throws InterruptedException {
        offer(e);
    }

    public boolean offer(E e, long timeout, TimeUnit unit) throws InterruptedException {
        return this.offer(e);
    }

    public int remainingCapacity() {
        return Integer.MAX_VALUE;
    }

    public int drainTo(Collection<? super E> c) {
        int i = 0;
        E e;

        for (; (e=q.poll()) != null; i++)
            c.add(e);
        return i; 
    }

    public int drainTo(Collection<? super E> c, int maxElements) {
        int i = 0;
        E e;

        for (; i < maxElements && (e=q.poll()) != null; i++)
            c.add(e);
        return i; 
    }
}